@@ -110,7 +110,7 @@ function(x, ...)
 {
     nr <- x$nrow
     nc <- x$ncol
-    y <- matrix(vector(typeof(x$v), nr * nc), nr, nc)
+    y <- matrix(vector(typeof(x$v), prod(nr, nc)), nr, nc)
     y[cbind(x$i, x$j)] <- x$v
     dimnames(y) <- x$dimnames
     y
@@ -457,7 +457,7 @@ function(x, value)
         stop("invalid dim replacement value")
     nr <- x$nrow
     nc <- x$ncol
-    if(prod(value) != nr * nc)
+    if(prod(value) != prod(nr, nc))
         stop("invalid dim replacement value")
 
     pos <- nr * (x$j - 1L) + x$i - 1L
@@ -515,6 +515,31 @@ function(x, value)
     x
 }
 
+## For reuse in other functions (we want to mess up, too).
+.stm_as_subscript <- 
+function(x, d, safe = FALSE, ...) 
+{
+    if (!is.simple_triplet_matrix(x))
+	return(x)
+    if (!is.logical(x$v) || !identical(dim(x), d))
+	stop("Not implemented.")
+    ## need column-major order
+    k <- order(x$j, x$i)
+    if (any(diff(k < 0))) {
+	x$v <- x$v[k]
+	x$i <- x$i[k]
+	x$j <- x$j[k]
+    }
+    ## offer a choice
+    if (safe ||
+	log2(prod(dim(x))) > .Machine$double.digits)
+	cbind(x$i[x$v], x$j[x$v])
+    else
+	## need to use a double in expression to 
+	## get a result of type double
+	((x$j - 1) * x$nrow + x$i)[x$v]
+}
+
 `[.simple_triplet_matrix` <-
 function(x, i, j, drop = FALSE)
 {
@@ -534,25 +559,33 @@ function(x, i, j, drop = FALSE)
     nr <- x$nrow
     nc <- x$ncol
 
+    pd <- prod(nr, nc)
+
+    ## FIXME eventually, we should get rid of ill-conceived features 
+    ##       which need to expand to dense.
+    .disable <- pd > slam_options("max_dense")
+
     if(na == 2L) {
         ## Single index subscripting.
         ## Mimic subscripting matrices: no named argument handling in
         ## this case.
+
+	## FIXME mapping to numeric seems to be less inefficient.
+	i <- .stm_as_subscript(i, c(nr, nc))
         if(is.character(i))
             out <- vector(typeof(x$v))[rep.int(NA, length(i))]
         else if(!is.matrix(i)) {
             if(is.logical(i)) {
-		if(nr * nc > 16777216L)
+		if(.disable)
 		  stop("Logical vector subscripting disabled for this object.")
-                i <- which(rep(i, length.out = nr * nc))
+                i <- which(rep_len(i, pd))
 	    }
-            else if(!is.numeric(i))
+            else if(!is.numeric(unclass(i)))
                 stop(gettextf("Invalid subscript type: %s.",
                               typeof(i)),
                      domain = NA)
 	    else
-		## 52-bit safe
-		if(nr * nc > 4503599627370496)
+		if(log2(pd) > .Machine$double.digits)
 		  stop("Numeric vector subscripting disabled for this object.")
 	    ## Shortcut
 	    if(!length(i))
@@ -565,15 +598,15 @@ function(x, i, j, drop = FALSE)
                 i <- i[i > 0]
                 out <- vector(mode = typeof(x$v), length = length(i))
 		if(length(out)) {
-		    is.na(i) <- i > nr * nc
+		    is.na(i) <- i > pd
 		    is.na(out) <- is.na(i)
-		    i <- match(i, (x$j - 1L) * nr + x$i, 0L)
+		    i <- match(i, (x$j - 1) * nr + x$i, 0L)
 		    out[i > 0L] <- x$v[i]
 		}
             } else if(!any(is.na(i)) && all(i <= 0)) {
-		if(nr * nc > 16777216L)
+		if(.disable)
 		  stop("Negative vector subscripting disabled for this object.")
-                out <- vector(mode = typeof(x$v), nr * nc)
+                out <- vector(mode = typeof(x$v), pd)
                 out[(x$j - 1L) * nr + x$i] <- x$v
                 out <- out[i]
             }
@@ -584,21 +617,21 @@ function(x, i, j, drop = FALSE)
 	    if(!nrow(i))
 		return(vector(mode = typeof(x$v), length = 0L))
 	    ## Ignore dimensions
-	    if(ncol(i) != 2L)
+	    if(ncol(i) != 2L || !is.numeric(i))
 		return(do.call("[.simple_triplet_matrix",
 			       list(x = x, as.vector(i))))
 
-            ## Note that negative values are not allowed in a matrix
-            ## subscript.
 	    if(is.double(i))
 		i <- trunc(i)
-            if(any(i < 0, na.rm = TRUE))
-                stop("Invalid subscript.")
             ## Rows containing zero indices can be dropped.
             ## Rows with NA indices should give NA (at least for
             ## non-recursive x).
 	    k <- .Call(R_all_row, i > 0, FALSE)
             i <- i[k, ,drop = FALSE]
+            ## Note that negative values are not allowed in a matrix
+            ## subscript.
+            if(any(i < 0, na.rm = TRUE))
+                stop("Invalid subscript.")
             out <- vector(mode = typeof(x$v), length = nrow(i))
 	    if(length(out)) {
 		if (any(i > rep(c(nr, nc), each = nrow(i)), na.rm = TRUE))
@@ -627,7 +660,7 @@ function(x, i, j, drop = FALSE)
 		stop("NA indices not allowed.")
             pi <- seq_len(nr)
             if(is.logical(i)) {
-                i <- rep(i, length.out = nr)
+                i <- rep_len(i, nr)
                 nr <- sum(i)
                 pos <- i[x$i]
             } else {
@@ -665,7 +698,7 @@ function(x, i, j, drop = FALSE)
 		stop("NA indices not allowed.")
             pj <- seq_len(nc)
             if(is.logical(j)) {
-                j <- rep(j, length.out = nc)
+                j <- rep_len(j, nc)
                 nc <- sum(j)
                 pos <- if(is.null(pos))
                     j[x$j]
@@ -850,7 +883,7 @@ function(..., recursive = FALSE)
     args[ind] <-
         lapply(args[ind],
                function(x) {
-                   y <- vector(typeof(x$v), x$nrow * x$ncol)
+                   y <- vector(typeof(x$v), prod(x$nrow, x$ncol))
                    y[x$i + (x$j - 1L) * x$nrow] <- x$v
                    y
                })
@@ -904,7 +937,7 @@ function(nrow, ncol = nrow, mode = "double")
 simple_triplet_diag_matrix <-
 function(v, nrow = length(v))
 {
-    v <- rep(v, length.out = nrow)
+    v <- rep_len(v, nrow)
     i <- seq_len(nrow)
     simple_triplet_matrix(i, i, v, nrow, nrow)
 }
